#include "program_version.h"
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <hardware/btn-hardware.h>
#include <terminal/btn-terminal.h>
#include <terminal/btn-cmds.h>
#include <terminal/nut-cmds.h>
#include <terminal/eeprom_cmds.h>
#include <led/btn-led.h>
#include <wsnlab/wsnlab.h>
#include <sys/timer.h>
#include <cc/bmac.h>
#include <cc/ccc.h>
#include <cc/cc1000.h> // for rf_power
#include <sys/event.h>

// ----------------------------------------------------------------------------
// #define constants
// ----------------------------------------------------------------------------

#define PROJECT_NAME "t2-example"
#define UART_SPEED   57600

#define CHAT_PACKET 0x01
#define PING_PACKET 0x02
#define PONG_PACKET 0x03

// ----------------------------------------------------------------------------
// variables
// ----------------------------------------------------------------------------

static u_short pong_addr;
static HANDLE pong_event;
 
// ----------------------------------------------------------------------------
// helper functions
// ----------------------------------------------------------------------------

/**
 * Initializes the BTnode.
 */
static void btnode_init(void) 
{    
    /* initialize the btnode hardware */
    btn_hardware_init();
    btn_led_init(0);

    char prompt[TERMINAL_MAX_PROMPT_LENGTH];
    u_short mac_address = get_MAC_address_from_EEPROM();
    
    if ( mac_address == 0 )
    {
    	//printf( "Error! No MAC-address in EEPROM!\n" ); // terminal not inited
    	mac_address = 0xEEEE;    
    }
    
    //initialise random number generator
    srand( mac_address );
        
    sprintf( prompt, "[%s@%02X:%02X]$", PROJECT_NAME, mac_address >> 8, 
    	mac_address & 0xFF );
    
    term_init(UART_SPEED, prompt);     
    
    int res = bmac_init(mac_address);
    if (res != 0)
    {
    	printf( "bmac_init() failed\n" );
    }
    bmac_enable_led( 1 );
    res = ccc_init(&bmac_interface);
    if (res != 0)
    {
    	printf( "ccc_init() failed\n" );
    }
}

int min( int first, int second )
{
	return (first > second ? second : first);
}

int max( int first, int second )
{
	return (first < second ? second : first);
}


// ----------------------------------------------------------------------------
// packet handling
// ----------------------------------------------------------------------------

/**
 * packet handler for ping packets
 */
void ping_packet_handler(ccc_packet_t *pkt)
{
 	printf( "Received ping ('%s') from 0x%04X...\n", (char*) (pkt->data), pkt->src);
 	
 	// remember last seen address
 	pong_addr = pkt->src;
 	
 	// schedule a pong packet
 	NutEventPost( &pong_event );
}

/**
 * packet handler for pong packets
 */
void pong_packet_handler(ccc_packet_t *pkt)
{
 	// print received pong packet
 	printf( "Received pong ('%s') from 0x%04X...\n", (char*) (pkt->data), pkt->src );
}

/**
 * packet handler for chat packets
 */
void chat_packet_handler(ccc_packet_t *pkt)
{
	// add '\0' to the end of the received packet to prevent buffer overflows
	static u_char buffer[256];
	memcpy( buffer, pkt->data, min( pkt->length, 256 ) );
	buffer[min( pkt->length, 255 )] = '\0';
	
	if ( pkt->dst == 0xFFFF )
	{
		// this packet went to all
		printf( "0x%04X says: %s\n", pkt->src, buffer );		
	}
	else
	{
		// this packet was addressed only to me 
		printf( "0x%04X whispers: %s\n", pkt->src, buffer ); 
	}
}

static void register_packet_handlers(void)
{	
	ccc_register_packet_handler(PING_PACKET, ping_packet_handler); 
	ccc_register_packet_handler(PONG_PACKET, pong_packet_handler);
	ccc_register_packet_handler(CHAT_PACKET, chat_packet_handler);
}


THREAD(PongThread, args)
{
	while( 1 )
	{
		NutEventWait( &pong_event, NUT_WAIT_INFINITE );
		printf( "Ponging 0x%04X...\n", pong_addr );
		char* content = "PONG";
		ccc_send_packet( pong_addr, PONG_PACKET, (u_char*) content, strlen(content) + 1 );
	}
}

// ----------------------------------------------------------------------------
// terminal commands
// ----------------------------------------------------------------------------

void ping_cmd(char* args)
{
	u_int dest;
	
	if ( sscanf( args, "%x", &dest ) != 1 )
	{
		printf( "usage: ping <dest>\n" );
	}
	else
	{
		printf( "Pinging 0x%04X...\n", (u_short) dest );	
		char* content = "PING";
		ccc_send_packet( (u_short) dest, PING_PACKET, (u_char*) content, strlen(content) + 1 );
	}
}

void say_cmd(char* args)
{
	if ( strlen( args ) == 0 )
	{
		printf( "usage: say <message>\n" );
	}
	else
	{
		printf( "You shout out loud: %s\n", args );
		ccc_send_packet( 0xFFFF, CHAT_PACKET, (u_char*) args, strlen(args) + 1 );
	} 
}

void whisper_cmd(char* args)
{
	u_int dest;
	char* string_ptr = args;
	char* end_ptr    = args + max(strlen(args) - 1, 0);

	// remove whitespaces at the end of the input
	while (end_ptr != args && *end_ptr == ' ')
	{
		*end_ptr = '\0';
		end_ptr--;
	}
	
	// remove whitespace before the first token	
	while (*string_ptr != '\0' && *string_ptr == ' ')
	{
		string_ptr++;
	}
	// skip first token
	while (*string_ptr != '\0' && *string_ptr != ' ')
	{
		string_ptr++;
	}
	// remove whitespace before second token
	while (*string_ptr != '\0' && *string_ptr == ' ')
	{
		string_ptr++;
	}
	// (string_ptr should now point to the second token)
	
	if ((*string_ptr == '\0') || 
	    sscanf( args, "%x", &dest ) != 1) 
	{
		printf( "usage: whisper <dest> <message>\n" );
	}	
	else
	{
		printf( "You whisper to 0x%04X: \"%s\"\n", (u_short) dest, string_ptr );	
		ccc_send_packet( (u_short) dest, CHAT_PACKET, (u_char*) string_ptr, strlen(string_ptr) + 1 );
	}
}

void rfpower_cmd(char *arg)
{
	u_short num;
	u_char num2;
	
    if (( sscanf(arg, "%hu", &num) != 1 ) || num > 255 ) 
    {
        printf("usage: rfpower <0..255>\n");
        cc1000_get_RF_power( &num2 );
        printf( "Current RF power level is %u.\n", num2 );
        return;
    }
    
    printf( "Setting RF power to %u...\n", num );
    cc1000_set_RF_power( num );
}

/**
 * Registers all supported commands.
 */
static void register_cmds(void)
{
	nut_cmds_register_cmds();
    btn_cmds_register_cmds();
    wsnlab_cmds_register_cmds();
    btn_terminal_register_cmd("say", say_cmd);
    btn_terminal_register_cmd("whisper", whisper_cmd);
	btn_terminal_register_cmd("ping", ping_cmd);
	btn_terminal_register_cmd("rfpower", rfpower_cmd);    
    eeprom_cmds_register_cmds();
}
  
int main(void)
{
	btnode_init();
    
    DEBUGB("\n# ----------------------------------------");
    DEBUGB("\n# " PROJECT_NAME " version " PROGRAM_VERSION "\n");
    DEBUGB("# ----------------------------------------\n");
    NutSleep(200);
 
    // register cmds
	register_cmds();
	
	// register packet handlers
	register_packet_handlers();
	
	NutThreadCreate("PongThread", PongThread, 0, 192);	
	
    // start terminal
    btn_terminal_run(BTN_TERMINAL_NOFORK, 1024);
    
    // we should never reach this point
    return 1;
}
